"""
A collection of wrappers
"""
from showspec_splat1d import splat_1d
from fit_gaussians_to_simple_spectra import fit_gaussians_to_simple_spectra
import fitnh3
import fith2co
from cube_fit import cube_fit
from load_IRAF_multispec import load_IRAF_multispec
